# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.api.requests import affinity


def listComponents(limit=9999):
    return affinity.post("/component/list", json={"limit": limit})["list"]


def shareComponent(componentId, userId, name):
    return affinity.post(
        "/component/share",
        json={"id": componentId, "targetUserId": userId, "name": name},
    )



# coding=utf-8
from __future__ import absolute_import, print_function

import os

environ = os.environ


def lazyget(
    key, default=None, required=False, type=str
):  # pylint: disable=used-before-assignment
    return property(
        lambda self: get(key, default=default, required=required, type=type)
    )


def get(key, default=None, required=False, type=str):
    if key not in environ:
        if required:
            raise Exception(f"No such env: {key}")
        return default
    value = environ[key]
    try:
        return type(value)
    except Exception:
        raise Exception(f"EnvTypeErr: ({key}) {value} except {getTypeName(type)}")


def getTypeName(type):
    name = getattr(type, "name", None) or getattr(type, "__name__", None)
    if not name:
        raise Exception(f"Unknown env type: {type}")
    return name


def update(*args, **kwargs):
    return environ.update(*args, **kwargs)


def str(value):
    return str(value)


def int(value):
    return int(value)


def float(value):
    return float(value)


def bool(value):
    return value in ("true", "True")



from flask import Flask, request  # pylint: disable=import-error

import suanpan
from suanpan import asyncio, g
from suanpan.app import app
from suanpan.app.arguments import String
from suanpan.log import logger
from suanpan.utils import json

if g.debug:
    web = Flask(__name__)

    @web.route("/", methods=["POST"])
    def testWeb():
        with asyncio.lock(thread=True):
            app.args.update(request.json)
        logger.debug(f"App args updated: {app.args}")
        return json.dumps(app.args)

    @app.afterInit
    def startWeb(_):
        asyncio.run(web.run, kwds={"host": "0.0.0.0", "port": 8000}, thread=True)


@app.trigger(interval=10)
@app.trigger.output(String(key="outputData1"))
@app.trigger.param(String(key="param1"))
def test(context):
    args = context.args
    with asyncio.lock(thread=True):
        logger.debug(f"Trigger: {args}")


if __name__ == "__main__":
    suanpan.run(app)



# coding=utf-8
from __future__ import absolute_import, print_function

import contextlib
import copy
import itertools
import time
import traceback
import uuid

from suanpan import asyncio, error, g, runtime, utils
from suanpan.arguments import Bool, BoolOrInt, Float, Int, String
from suanpan.components import Arguments, Component
from suanpan.dw import dw
from suanpan.interfaces import (
    HasCallHooks,
    HasDevMode,
    HasExitHooks,
    HasInitHooks,
    HasLogger,
    HasTriggerHooks,
)
from suanpan.interfaces.optional import HasBaseServices
from suanpan.log import logger
from suanpan.mq import mq
from suanpan.mstorage import mstorage
from suanpan.objects import Context
from suanpan.storage import storage
from suanpan.utils import json


class Handler(Component):
    def run(self, steamObj, message, *arg, **kwargs):
        self.callBeforeInitHooks()
        context = self.init(steamObj, message)
        self.callAfterInitHooks(context)

        self.callBeforeCallHooks(context)
        if not self.runFunc:
            raise error.ComponentHandlerNotSet()
        if not callable(self.runFunc):
            raise error.ComponentHandlerNotCallable(self.name)
        results = self.runFunc(steamObj, context, *arg, **kwargs)
        self.callAfterCallHooks(context)

        self.callBeforeSaveHooks(context)
        outputs = self.save(results, message=message)
        self.callAfterSaveHooks(context)

        self.callBeforeCleanHooks(context)
        self.clean()
        self.callAfterCleanHooks()

        return outputs

    def beforeInit(self):
        logger.debug(f"Handler {self.name} starting...")

    def init(self, steamObj, message):
        self.argsDict = self.getArgsDict(message)
        context = self._getContext(message)
        args = self.loadComponentArguments(self.argsDict)
        args.update(steamObj.args)
        context.update(args=args)
        return context

    def load(self, args, argsDict=None):
        logger.info("Loading Arguments:")
        argsDict = argsDict or self.argsDict
        argsDict = {
            f"{arg.key}": argsDict.get(f"in{i+1}")
            for i, arg in enumerate(args)
            if argsDict.get(f"in{i+1}") is not None
        }
        args = self.loadFormatArguments(args, argsDict)
        return Arguments.froms(*args)

    def loadComponentArguments(self, argsDict=None):
        logger.info("Loading Component Arguments:")
        argsDict = argsDict or self.argsDict
        arguments = copy.deepcopy(self.getArguments(exclude="outputs"))
        argsDict = {
            f"{arg.key}": argsDict.get(f"in{i+1}")
            for i, arg in enumerate(arguments)
            if argsDict.get(f"in{i+1}") is not None
        }
        argsDict = argsDict or self.argsDict
        args = self.loadFormatArguments(arguments, argsDict)
        return Arguments.froms(*args)

    def getArgsDict(self, message):
        return message or {}

    @contextlib.contextmanager
    def context(self, message):
        yield Context.froms(message=message)

    def save(self, results, args=None, message=None):
        args = args or copy.deepcopy(self.getArguments(include="outputs"))
        shortRequestID = self.shortenRequestID(message["id"])
        outputArgsDict = {
            arg.key: arg.getOutputTmpValue(
                "studio", g.userId, g.appId, shortRequestID, g.nodeId, f"out{i+1}"
            )
            for i, arg in enumerate(args)
        }
        args = self.loadCleanArguments(args, outputArgsDict)
        return self.saveOutputs(args, results)

    def afterSave(self, context):  # pylint: disable=unused-argument
        logger.debug(f"Handler {self.name} done.")

    def saveOutputs(self, args, results):
        if results is not None:
            outputs = super(Handler, self).saveOutputs(args, results)
            outputs = self.formatAsOuts(args, outputs)
            outputs = self.stringifyOuts(outputs)
            return outputs
        return None

    def formatAsOuts(self, args, results):
        return {
            f"out{i+1}": self.getArgumentValueFromDict(results, arg)
            for i, arg in enumerate(args)
        }

    def stringifyOuts(self, outs):
        return {k: str(v) for k, v in outs.items() if v is not None}

    def shortenRequestID(self, requestID):
        return requestID.replace("-", "")


class StreamBase(HasBaseServices, HasLogger, HasDevMode, HasInitHooks, HasExitHooks):

    DEFAULT_LOGGER_MAX_LENGTH = 120
    DEFAULT_MESSAGE = {}
    DEFAULT_STREAM_CALL = "call"
    STREAM_ARGUMENTS = [
        String("stream-recv-queue", default=f"mq-{g.nodeId}"),
        BoolOrInt("stream-recv-queue-block", default=60000),
        Float("stream-recv-queue-delay", default=0),
        Int("stream-recv-queue-max-length", default=1000),
        Bool("stream-recv-queue-trim-immediately", default=False),
        Bool("stream-recv-queue-retry", default=False),
        Int("stream-recv-queue-retry-max-count", default=100),
        Float("stream-recv-queue-retry-timeout", default=1.0),
        Int("stream-recv-queue-retry-max-times", default=3),
        String("stream-send-queue", default="mq-master"),
        Int("stream-send-queue-max-length", default=1000),
        Bool("stream-send-queue-trim-immediately", default=False),
    ]

    def __init__(self):
        super(StreamBase, self).__init__()
        self.argsDict = {}
        self.args = None
        self.options = {}

    def beforeInit(self):
        logger.logDebugInfo()
        logger.debug(f"Stream {self.name} starting...")

    def init(self, *args, **kwargs):
        self.argsDict = self.getArgsDict(*args, **kwargs)
        self.args = self.loadGlobalArguments(self.argsDict)
        self.options = self.getOptions(self.args)
        self.setBaseServices(self.args)
        self.args.update(self.loadComponentArguments(self.argsDict))
        return Context.froms(args=self.args)

    def loadGlobalArguments(self, argsDict=None):
        logger.info("Loading Global Arguments:")
        argsDict = argsDict or self.argsDict
        args = self.loadFormatArguments(self.getGlobalArguments(), argsDict)
        return Arguments.froms(*args)

    def loadComponentArguments(self, argsDict=None):
        logger.info("Loading Components Arguments:")
        argsDict = argsDict or self.argsDict
        args = self.loadFormatArguments(self.getComponentArguments(), argsDict)
        return Arguments.froms(*args)

    def getGlobalArguments(self, *args, **kwargs):
        arguments = super(StreamBase, self).getGlobalArguments(*args, **kwargs)
        return arguments + self.STREAM_ARGUMENTS

    def generateRequestId(self):
        return uuid.uuid4().hex

    def generateMessage(self, **kwargs):
        message = {}
        message.update(self.DEFAULT_MESSAGE, **kwargs)
        message.setdefault("type", self.DEFAULT_STREAM_CALL)
        message["id"] = self.generateRequestId()
        return message

    def formatMessage(self, message, msg, costTime=None):
        msgs = [message["id"], message.get("type", self.DEFAULT_STREAM_CALL), msg]
        if costTime is not None:
            msgs.insert(-1, f"{costTime}s")
        return " - ".join(msgs)

    def streamCall(self, message, *args, **kwargs):
        logger.info(self.formatMessage(message, msg="Start"))
        startTime = time.time()
        try:
            handler = self.getHandler(message)
            outputs = handler.run(self, message, *args, **kwargs) or {}
            endTime = time.time()
            costTime = round(endTime - startTime, 3)
            logger.info(self.formatMessage(message, msg="Done", costTime=costTime))
            if outputs:
                self.sendSuccessMessage(message, outputs)
        except Exception:  # pylint: disable=broad-except
            tracebackInfo = traceback.format_exc()
            endTime = time.time()
            costTime = round(endTime - startTime, 3)
            logger.error(
                self.formatMessage(message, msg=tracebackInfo, costTime=costTime)
            )
            self.sendFailureMessage(message, tracebackInfo)

    def handlerCallback(self, *args, **kwargs):
        return self.streamCall(self.generateMessage(), *args, **kwargs)

    def getHandler(self, message):
        streamCall = message.get("type", self.DEFAULT_STREAM_CALL)
        handler = getattr(self, streamCall, None)
        if not handler or not isinstance(handler, Handler):
            raise error.StreamError(f"Unknown stream handler {self.name}.{streamCall}")
        return handler

    @runtime.globalrun
    def start(self, *args, **kwargs):
        self.callBeforeInitHooks()
        context = self.init(*args, **kwargs)
        self.callAfterInitHooks(context)
        self.registerBeforeExitHooks(context)
        self.run()

    def run(self):
        self.startCallLoop()

    def startCallLoop(self):
        if self.options["recvQueueRetry"]:
            self.retryPendingMessages()
        for message in self.subscribe():
            self.handleMessage(message.get("data", {}))

    @runtime.saferun
    def handleMessage(self, message):
        self.streamCall(message)

    def setDefaultMessageType(self, message):
        message["data"].setdefault("type", self.DEFAULT_STREAM_CALL)
        return message

    def getMessageExtraData(self, message):
        extra = message["data"].get("extra")
        extra = json.loads(extra) if extra else {}
        message["data"].update(extra=extra)
        return message

    def getOptions(self, args):
        return self.defaultArgumentsFormat(args, self.STREAM_ARGUMENTS)

    def subscribe(self, **kwargs):
        for message in mq.subscribeQueue(
            self.options["recvQueue"],
            group=g.nodeGroup,
            consumer=g.nodeId,
            block=self.options["recvQueueBlock"],
            delay=self.options["recvQueueDelay"],
            **kwargs,
        ):
            message = self.setDefaultMessageType(message)
            message = self.getMessageExtraData(message)
            yield message

    def recv(self, **kwargs):
        return mq.recvMessages(
            self.options["recvQueue"], group=g.nodeId, consumer=self.name, **kwargs,
        )

    def _send(self, message, data, queue=None, extra=None):
        queue = queue or self.options["sendQueue"]
        message.setdefault("extra", {})
        message["extra"].update(extra or {})
        data = {
            "node_id": g.nodeId,
            "request_id": message["id"],
            "type": self.DEFAULT_STREAM_CALL,
            "extra": json.dumps(message["extra"]),
            **data,
        }
        logger.debug(
            utils.shorten(f"Send to `{queue}`: {data}", self.DEFAULT_LOGGER_MAX_LENGTH)
        )
        return mq.sendMessage(
            queue,
            data,
            maxlen=self.options["sendQueueMaxLength"],
            trimImmediately=self.options["sendQueueTrimImmediately"],
        )

    def sendSuccessMessage(self, message, data, queue=None, extra=None):
        if not all(key.startswith("out") for key in data):
            raise error.StreamError(
                "Success Message data only accept keys starts with 'out'"
            )
        data = {key: value for key, value in data.items() if value is not None}
        data.update(success="true")
        return self._send(message, data, queue=queue, extra=extra)

    def sendFailureMessage(self, message, msg, queue=None, extra=None):
        if not isinstance(msg, str):
            raise error.StreamError("Failure Message msg only accept string")
        data = {"msg": msg, "success": "false"}
        return self._send(message, data, queue=queue, extra=extra)

    def send(self, results, queue=None, message=None, args=None, extra=None, **_):
        message = message or self.generateMessage()
        outputs = self.getHandler(message).save(results, args=args, message=message)
        return self.sendSuccessMessage(message, outputs, queue=queue, extra=extra)

    def sendError(self, msg, queue=None, message=None, extra=None):
        message = message or self.generateMessage()
        return self.sendFailureMessage(message, msg, queue=queue, extra=extra)

    def sendMissionMessage(self, message, data, queue=None, extra=None):
        if not all(key.startswith("in") for key in data):
            raise error.StreamError(
                "Mission Message data only accept keys starts with 'in'"
            )
        data = {key: value for key, value in data.items() if value is not None}
        data.update(id=message["id"])
        return self._send(message, data, queue=queue, extra=extra)

    def retryPendingMessages(self, **kwargs):
        return mq.retryPendingMessages(
            self.options["recvQueue"],
            group=g.nodeGroup,
            consumer=g.nodeId,
            count=self.options["recvQueueRetryMaxCount"],
            maxTimes=self.options["recvQueueRetryMaxTimes"],
            timeout=self.options["recvQueueRetryTimeout"],
            maxlen=self.options["recvQueueMaxLength"],
            trimImmediately=self.options["recvQueueTrimImmediately"],
            **kwargs,
        )

    def keysAllIn(self, keys, kset):
        return len(set(keys) - set(kset)) == 0


class Stream(StreamBase):
    INTERVAL = 0
    TRIGGER_ARGUMENTS = [Float("triggerInterval", default=0)]
    DEFAULT_TRIGGER_CALL = "trigger"

    def getGlobalArguments(self, *args, **kwargs):
        arguments = super(Stream, self).getGlobalArguments(*args, **kwargs)
        return arguments + self.TRIGGER_ARGUMENTS

    def _list(self, data):
        if isinstance(data, (tuple, list)):
            return data
        if data is None:
            return []
        return [data]

    @property
    def interval(self):
        _interval = getattr(self, "_interval", None)
        if _interval is not None:
            return _interval
        return self.args.triggerInterval or self.INTERVAL

    @interval.setter
    def interval(self, value):
        if not isinstance(value, (int, float)):
            raise error.StreamError("Interval must be int or float")
        setattr(self, "_interval", value)

    def loop(self):
        while True:
            yield
            time.sleep(self.interval)

    def triggerCall(self, *args, **kwarags):
        return self.streamCall(
            self.generateMessage(type=self.DEFAULT_TRIGGER_CALL), *args, **kwarags
        )

    def startTriggerLoop(self):
        for data in self.loop():
            self.triggerCall(*self._list(data))

    def getInitialRunners(self):
        runners = []
        stream = getattr(self, self.DEFAULT_STREAM_CALL, None)
        if stream and stream.runFunc:
            runners.append(self.startCallLoop)
        trigger = getattr(self, self.DEFAULT_TRIGGER_CALL, None)
        if trigger and trigger.runFunc:
            runners.append(self.startTriggerLoop)
        return runners

    def startOnlyRunner(self, runners):
        runners[0]()

    def startMultiRunners(self, runners):
        asyncio.wait(asyncio.run(runners, workers=len(runners), thread=True))

    def run(self):
        runners = self.getInitialRunners()
        if not runners:
            raise error.StreamError("Stream can not start: No Method implemented!")

        if len(runners) == 1:
            self.startOnlyRunner(runners)
        else:
            self.startMultiRunners(runners)

    def triggerSend(self, results, queue=None, message=None, args=None, extra=None):
        message = message or self.generateMessage(type=self.DEFAULT_TRIGGER_CALL)
        return self.send(results, queue=queue, message=message, args=args, extra=extra)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.objects import HasName


class BaseApp(HasName):
    def __call__(self, *args, **kwargs):
        raise NotImplementedError("Method not implemented!")

    def start(self, *args, **kwargs):
        raise NotImplementedError("Method not implemented!")

    @property
    def trigger(self):
        raise NotImplementedError(f"{self.name} not support trigger")

    def input(self, argument):
        raise NotImplementedError("Method not implemented!")

    def output(self, argument):
        raise NotImplementedError("Method not implemented!")

    def param(self, argument):
        raise NotImplementedError("Method not implemented!")

    def column(self, argument):
        raise NotImplementedError("Method not implemented!")

    def beforeInit(self, hook):
        raise NotImplementedError("Method not implemented!")

    def afterInit(self, hook):
        raise NotImplementedError("Method not implemented!")

    def beforeCall(self, hook):
        raise NotImplementedError("Method not implemented!")

    def afterCall(self, hook):
        raise NotImplementedError("Method not implemented!")

    def beforeExit(self, hook):
        raise NotImplementedError("Method not implemented!")

    def load(self, *args, **kwargs):
        raise NotImplementedError(f"{self.name} not support load")

    def save(self, *args, **kwargs):
        raise NotImplementedError(f"{self.name} not support save")

    def send(self, *args, **kwargs):
        raise NotImplementedError(f"{self.name} not support send")

    @property
    def args(self):
        raise NotImplementedError(f"{self.name} not support args")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import error
from suanpan.app.base import BaseApp
from suanpan.model.arguments import HotReloadModel
from suanpan.stream import Handler, Stream
from suanpan.utils import functional

COMPONENT_ARGUMENT_CLASSES = (HotReloadModel,)


class TriggerApp(BaseApp):
    def __init__(self, stream, *args, **kwargs):
        super(TriggerApp, self).__init__(*args, **kwargs)
        self.stream = stream
        self.stream.trigger = Handler()

    def __call__(self, funcOrApp):
        if not isinstance(funcOrApp, BaseApp):
            self.stream.trigger.use(functional.instancemethod(funcOrApp))
        return self

    def interval(self, _interval):
        self.stream.INTERVAL = _interval
        return self

    def loop(self, _loop):
        self.stream.loop = functional.instancemethod(_loop)
        return self

    def input(self, argument):
        if isinstance(argument, COMPONENT_ARGUMENT_CLASSES):
            self.stream.ARGUMENTS.append(argument)
        else:
            self.stream.trigger.input(argument)
        return self

    def output(self, argument):
        if isinstance(argument, HotReloadModel):
            raise error.AppError(f"{argument.name} can't be set as output!")
        self.stream.trigger.output(argument)
        return self

    def param(self, argument):
        self.stream.ARGUMENTS.append(argument)
        return self

    def column(self, argument):
        self.stream.ARGUMENTS.append(argument)
        return self

    def beforeInit(self, hook):
        self.stream.addBeforeInitHooks(hook)
        return hook

    def afterInit(self, hook):
        self.stream.addAfterInitHooks(hook)
        return hook

    def beforeCall(self, hook):
        self.stream.trigger.addBeforeCallHooks(hook)
        return hook

    def afterCall(self, hook):
        self.stream.trigger.addAfterCallHooks(hook)
        return hook

    def beforeExit(self, hook):
        self.stream.addBeforeExitHooks(hook)
        return hook

    def load(self, *args, **kwargs):
        return self.stream.trigger.load(*args, **kwargs)

    def send(self, *args, **kwargs):
        return self.stream.triggerSend(*args, **kwargs)


class StreamApp(BaseApp):
    def __init__(self, *args, **kwargs):
        super(StreamApp, self).__init__(*args, **kwargs)
        self.stream = Stream()
        self.stream.call = Handler()
        self._trigger = TriggerApp(self.stream)

    def __call__(self, funcOrApp):
        if not isinstance(funcOrApp, BaseApp):
            self.stream.call.use(functional.instancemethod(funcOrApp))
        return self

    def start(self, *args, **kwargs):
        return self.stream.start(*args, **kwargs)

    def input(self, argument):
        if isinstance(argument, COMPONENT_ARGUMENT_CLASSES):
            self.stream.ARGUMENTS.append(argument)
        else:
            self.stream.call.input(argument)
        return self

    def output(self, argument):
        if isinstance(argument, HotReloadModel):
            raise error.AppError(f"{argument.name} can't be set as output!")
        self.stream.call.output(argument)
        return self

    def param(self, argument):
        self.stream.ARGUMENTS.append(argument)
        return self

    def column(self, argument):
        self.stream.ARGUMENTS.append(argument)
        return self

    def beforeInit(self, hook):
        self.stream.addBeforeInitHooks(hook)
        return hook

    def afterInit(self, hook):
        self.stream.addAfterInitHooks(hook)
        return hook

    def beforeCall(self, hook):
        self.stream.call.addBeforeCallHooks(hook)
        return hook

    def afterCall(self, hook):
        self.stream.call.addAfterCallHooks(hook)
        return hook

    def beforeExit(self, hook):
        self.stream.addBeforeExitHooks(hook)
        return hook

    def load(self, *args, **kwargs):
        return self.stream.call.load(*args, **kwargs)

    def save(self, *args, **kwargs):
        return self.stream.send(*args, **kwargs)

    def send(self, *args, **kwargs):
        return self.stream.send(*args, **kwargs)

    @property
    def args(self):
        return self.stream.args

    @property
    def trigger(self):
        return self._trigger



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.arguments import String
from suanpan.storage import storage
from suanpan.tools import ToolComponent as tc


@tc.param(String(key="local", required=True))
@tc.param(String(key="remote", required=True))
def SPStorageDownload(context):
    args = context.args
    storage.download(args.remote, args.local)


if __name__ == "__main__":
    SPStorageDownload()  # pylint: disable=no-value-for-parameter



import pdb

from suanpan import graph

info = graph.Graph(appId="8885")
info.addConnection(
    sourceNodeId=info.getNodeByName("VS Code - Python (Service)").id,
    sourcePortId="out1",
    targetNodeId=info.getNodeByName("web output").id,
    targetPortId="in1",
)
info.update()
pdb.set_trace()
info.revert()
pdb.set_trace()
info.revert()



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.api.requests import affinity


def getAppGraph(appId):
    return affinity.get(f"/app/graph/{appId}")["data"]


def updateAppGraph(appId, graph):
    return affinity.post(f"/app/graph/{appId}", json={"graph": graph})


def revertAppGraph(appId):
    return affinity.post(f"/app/graph/{appId}/revert")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.objects import Context
from suanpan.utils import env

g = Context(
    # App
    debug=env.lazyget("SP_DEBUG", type=env.bool, default=False),
    appType=env.lazyget("SP_APP_TYPE", required=True),
    appParams=env.lazyget("SP_PARAM", default=""),
    nodeInfo=env.lazyget("SP_NODE_INFO", default=""),
    userId=env.lazyget("SP_USER_ID", required=True),
    appId=env.lazyget("SP_APP_ID", required=True),
    nodeId=env.lazyget("SP_NODE_ID", required=True),
    nodeGroup=env.lazyget("SP_NODE_GROUP", default="default"),
    # Api
    host=env.lazyget("SP_HOST", required=True),
    hostTls=env.lazyget("SP_HOST_TLS", type=env.bool, default=False),
    apiHost=env.lazyget("SP_API_HOST", required=True),
    apiHostTls=env.lazyget("SP_API_HOST_TLS", type=env.bool, default=False),
    affinity=env.lazyget("SP_AFFINITY"),
    accessKey=env.lazyget("SP_ACCESS_KEY", required=True),
    accessSecret=env.lazyget("SP_ACCESS_SECRET", required=True),
    userIdHeaderField=env.lazyget("SP_USER_ID_HEADER_FIELD", default="x-sp-user-id"),
    userSignatureHeaderField=env.lazyget(
        "SP_USER_SIGNATURE_HEADER_FIELD", default="x-sp-signature"
    ),
    userSignVersionHeaderField=env.lazyget(
        "SP_USER_SIGN_VERSION_HEADER_FIELD", default="x-sp-sign-version"
    ),
    # Screenshots
    screenshotsType=env.lazyget("SP_SCREENSHOTS_TYPE", default="index"),
    screenshotsPattern=env.lazyget("SP_SCREENSHOTS_PATTERN"),
    screenshotsStorageKey=env.lazyget("SP_SCREENSHOTS_STORAGE_KEY"),
    screenshotsThumbnailStorageKey=env.lazyget("SP_SCREENSHOTS_THUMBNAIL_STORAGE_KEY"),
)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.utils import json


class Variable(object):
    def __init__(self, mstorage, key):
        self.mstorage = mstorage
        self.key = key
        self._value = None

    @property
    def value(self):
        if self._value is None:
            self._value = self.get()
        return self._value

    def get(self):
        self._value = self.decode(self.mstorage.get(self.key))
        return self._value

    def set(self, value):
        self._value = value
        self.mstorage.set(self.key, self.encode(self._value))

    def delete(self):
        self.mstorage.delete(self.key)

    @classmethod
    def encode(cls, value):
        return value

    @classmethod
    def decode(cls, value):
        return value


class String(Variable):
    pass


class Int(Variable):
    @classmethod
    def encode(cls, value):
        return str(value)

    @classmethod
    def decode(cls, value):
        return int(value)


class Float(Variable):
    @classmethod
    def encode(cls, value):
        return str(value)

    @classmethod
    def decode(cls, value):
        return float(value)


class Json(Variable):
    @classmethod
    def encode(cls, value):
        return json.dumps(value)

    @classmethod
    def decode(cls, value):
        return json.loads(value)



# coding=utf-8
from __future__ import absolute_import, print_function

import multiprocessing
import multiprocessing.dummy

from suanpan import utils
from suanpan.utils import pbar as spbar

WORKERS = multiprocessing.cpu_count()


class GlobalPool(object):
    def __init__(self, workers=WORKERS):
        super(GlobalPool, self).__init__()
        self._workers = workers

    @property
    def workers(self):
        return self._workers

    @workers.setter
    def workers(self, value):
        self._workers = value

    @utils.lazyproperty
    def threads(self):
        return Pool(thread=True, workers=self.workers)

    @utils.lazyproperty
    def processes(self):
        return Pool(thread=False, workers=self.workers)

    def pool(self, thread=False):
        return self.threads if thread else self.processes

    def imap(self, func, iterable, chunksize=1, pbar=None, total=None, thread=False):
        return self.pool(thread=thread).imap(
            func, iterable, chunksize=chunksize, pbar=pbar, total=total
        )

    def map(self, func, iterable, chunksize=1, pbar=None, total=None, thread=False):
        return self.pool(thread=thread).map(
            func, iterable, chunksize=chunksize, pbar=pbar, total=total
        )

    def istarmap(
        self, func, iterable, chunksize=1, pbar=None, total=None, thread=False
    ):
        return self.pool(thread=thread).istarmap(
            func, iterable, chunksize=chunksize, pbar=pbar, total=total
        )

    def starmap(self, func, iterable, chunksize=1, pbar=None, total=None, thread=False):
        return self.pool(thread=thread).starmap(
            func, iterable, chunksize=chunksize, pbar=pbar, total=total
        )

    def run(self, funcs, args=(), kwds=None, thread=False, **kwargs):
        return self.pool(thread=thread).run(funcs, args=args, kwds=kwds, **kwargs)

    def lock(self, thread=False):
        return self.pool(thread=thread).lock()

    def wait(self, results):
        wait(results)


class StarmapRunner(object):
    def __init__(self, func):
        self.func = func

    def __call__(self, args):
        return self.func(*args)


class Pool(object):
    def __init__(self, workers=None, thread=False, join=False):
        super(Pool, self).__init__()
        self.workers = workers or WORKERS
        self.thread = thread
        self.join = join

    @utils.lazyproperty
    def pool(self):
        PoolClass = multiprocessing.dummy.Pool if self.thread else multiprocessing.Pool
        return PoolClass(processes=self.workers)

    def close(self):
        self.pool.close()
        if self.join:
            self.pool.join()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def imap(self, func, iterable, chunksize=1, pbar=None, total=None):
        iterable, total = spbar.getIterableLen(iterable, config=pbar, total=total)
        return spbar.one(
            self.pool.imap(func, iterable, chunksize=chunksize),
            config=pbar,
            total=total,
        )

    def map(self, func, iterable, chunksize=1, pbar=None, total=None):
        return list(
            self.imap(func, iterable, chunksize=chunksize, pbar=pbar, total=total)
        )

    def istarmap(self, func, iterable, chunksize=1, pbar=None, total=None):
        return self.imap(
            StarmapRunner(func), iterable, chunksize=chunksize, pbar=pbar, total=total
        )

    def starmap(self, func, iterable, chunksize=1, pbar=None, total=None):
        return list(
            self.istarmap(func, iterable, chunksize=chunksize, pbar=pbar, total=total)
        )

    def run(self, funcs, args=(), kwds=None, **kwargs):
        kwds = kwds or {}
        single = not isinstance(funcs, (list, tuple))
        funcs = [funcs] if single else funcs
        results = [
            self.pool.apply_async(func, args=args, kwds=kwds, **kwargs)
            for func in funcs
        ]
        return results[0] if single else results

    def lock(self):
        return multiprocessing.dummy.Lock() if self.thread else multiprocessing.Lock()

    def wait(self, results):
        wait(results)


pool = GlobalPool()


def GLOBAL_WORKERS(workers):
    pool.workers = workers


def imap(
    func, iterable, chunksize=1, pbar=None, total=None, thread=False, workers=None
):
    if workers is not None:
        with Pool(workers=workers, thread=thread) as _pool:
            return _pool.imap(
                func, iterable, chunksize=chunksize, pbar=pbar, total=total
            )
    return pool.imap(
        func, iterable, chunksize=chunksize, pbar=pbar, total=total, thread=thread
    )


def map(func, iterable, chunksize=1, pbar=None, total=None, thread=False, workers=None):
    if workers is not None:
        with Pool(workers=workers, thread=thread) as _pool:
            return _pool.map(
                func, iterable, chunksize=chunksize, pbar=pbar, total=total
            )
    return pool.map(
        func, iterable, chunksize=chunksize, pbar=pbar, total=total, thread=thread
    )


def istarmap(
    func, iterable, chunksize=1, pbar=None, total=None, thread=False, workers=None
):
    if workers is not None:
        with Pool(workers=workers, thread=thread) as _pool:
            return _pool.istarmap(
                func, iterable, chunksize=chunksize, pbar=pbar, total=total
            )
    return pool.istarmap(
        func, iterable, chunksize=chunksize, pbar=pbar, total=total, thread=thread
    )


def starmap(
    func, iterable, chunksize=1, pbar=None, total=None, thread=False, workers=None
):
    if workers is not None:
        with Pool(workers=workers, thread=thread) as _pool:
            return _pool.starmap(
                func, iterable, chunksize=chunksize, pbar=pbar, total=total
            )
    return pool.starmap(
        func, iterable, chunksize=chunksize, pbar=pbar, total=total, thread=thread
    )


def run(funcs, args=(), kwds=None, thread=False, workers=None):
    if workers is not None:
        workers = min(len(funcs), workers)
        with Pool(workers=workers, thread=thread) as _pool:
            return _pool.run(funcs, args=args, kwds=kwds)
    return pool.run(funcs, args=args, kwds=kwds, thread=thread)


def lock(thread=False):
    return pool.lock(thread=thread)


def wait(results):
    single = not isinstance(results, (list, tuple))
    results = [results] if single else results
    results = [r.get() for r in results]
    return results[0] if single else results



# coding=utf-8
from __future__ import absolute_import, print_function

import abc

from suanpan.mstorage.vars import Float, Int, Json, String
from suanpan.objects import HasName


class MStorage(HasName):
    @abc.abstractmethod
    def get(self, name, *args, **kwargs):
        pass

    @abc.abstractmethod
    def set(self, name, value, *args, **kwargs):
        pass

    @abc.abstractmethod
    def mget(self, name, *args, **kwargs):
        pass

    @abc.abstractmethod
    def mset(self, name, mapping, *args, **kwargs):
        pass

    def Int(self, key):
        return Int(self, key)

    def Float(self, key):
        return Float(self, key)

    def String(self, key):
        return String(self, key)

    def Json(self, key):
        return Json(self, key)



# coding=utf-8
from __future__ import absolute_import, print_function

import collections
import functools
import time


def merge(*dicts):
    def _merge(result, item):
        result.update(item)
        return result

    return functools.reduce(_merge, dicts, {})


def count(iterable):
    if hasattr(iterable, "__len__"):
        return len(iterable)

    d = collections.deque(enumerate(iterable, 1), maxlen=1)
    return d[0][0] if d else 0


def shorten(value, maxlen=80):
    string = str(value)
    return string[: maxlen - 3] + "..." if len(string) > maxlen else string


class FrequenceTimeLimiter(object):
    def __init__(self, seconds):
        self.frequence = seconds
        self.time = time.time()

    def ifEnable(self):
        newTime = time.time()
        enable = self.frequence < newTime - self.time
        self.time = newTime if enable else self.time
        return enable


def encode(data):
    if isinstance(data, str):
        data = data.encode()
    return data


def decode(data):
    if isinstance(data, bytes):
        data = data.decode()
    return data


def lazyproperty(func):
    attrname = "_lazy_" + func.__name__

    @property
    def _lazyproperty(self):
        if not hasattr(self, attrname):
            setattr(self, attrname, func(self))
        return getattr(self, attrname)

    return _lazyproperty



# coding=utf-8
from __future__ import absolute_import, print_function

import enum
import traceback

from suanpan import error, utils
from suanpan.log import logger
from suanpan.objects import Context
from suanpan.utils import json


@enum.unique
class SocketKeyword(enum.Enum):
    PING = "ping"
    PONG = "pong"


class BaseDevice(object):
    def __init__(self, name, url, **kwargs):
        self.name = name
        self.url = url
        self.current = Context(**kwargs)

        self.connect()

    def connect(self):
        raise NotImplementedError("Method not implemented!")

    def close(self):
        raise NotImplementedError("Method not implemented!")

    def send(self, data, target=None):
        target = target or self.current.message.get("from")
        if not target:
            raise error.DeviceError("No target specified")
        self._send(utils.encode(target), utils.encode(data))

    def sendBytes(self, data, target=None):
        self.send(data, target=target)

    def sendJson(self, data, target=None):
        self.send(json.dumps(data), target=target)

    def recv(self, decode=True):
        pongBytes = utils.encode(SocketKeyword.PONG)
        while True:
            _from, data = self._recv()
            if data == pongBytes:
                self.pong()
                continue

            _from = utils.decode(_from)
            if decode:
                data = data.decode()
            self.current.message = {"from": _from, "data": data}
            return data

    def recvBytes(self):
        return self.recv(decode=False)

    def recvJson(self):
        return json.loads(self.recv(decode=True))

    def request(self, data, target=None, decode=True):
        self.send(data, target)
        return self.recv(decode=decode)

    def requestBytes(self, data, target=None):
        self.sendBytes(data, target)
        return self.recvBytes()

    def requestJson(self, data, target=None):
        self.sendJson(data, target)
        return self.recvJson()

    def _subscribe(self, _recv, *args, **kwargs):
        while True:
            try:
                yield self.recv(*args, **kwargs)
            except Exception:  # pylint: disable=broad-except
                logger.warning("Error in subscribe messages:")
                logger.warning(traceback.format_exc())

    def subscribe(self, decode=True):
        yield from self._subscribe(self.recv, decode=decode)

    def subscribeBytes(self):
        yield from self._subscribe(self.recvBytes)

    def subscribeJson(self):
        yield from self._subscribe(self.recvJson)

    def ping(self, target):
        result = self.request(SocketKeyword.PING, target, decode=True)
        if result != SocketKeyword.PONG:
            raise error.DeviceError(f"Ping error: {result}")

    def pong(self):
        self.send(SocketKeyword.PONG)

    def _send(self, data, target):
        raise NotImplementedError("Method not implemented!")

    def _recv(self):
        raise NotImplementedError("Method not implemented!")
